__title__ = 'mano'
__description__ = 'Python API for Beiwe Research Platform'
__url__ = 'https://www.neuroinfo.org'
__version__ = '0.5.1'
__author__ = 'Neuroinformatics Research Group'
__author_email__ = 'info@neuroinfo.org'

